"""Vorze script module."""
import math
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import ClassVar
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

from buttplug.core import LinearSubcommand
from buttplug.core import RotateSubcommand
from buttplug.core import SpeedSubcommand
from loguru import logger

from .base import BaseCommand
from .base import LinearPositionCommand
from .base import RotateCommand
from .base import VibrateCommand


_T = TypeVar("_T", bound="BaseVorzeCommand")


class BaseVorzeCommand(BaseCommand):
    """Base Vorze script command.

    Supports native serialization to Vorze CSV and LPEG VCSX.
    """

    @abstractmethod
    def to_csv(self) -> Iterable[Any]:
        """Return Vorze CSV row data for this command."""

    @classmethod
    @abstractmethod
    def from_csv(cls: Type[_T], row: Iterable[Any]) -> _T:
        """Construct command from a Vorze CSV row.

        Arguments:
            row: CSV row data.
        """

    @abstractmethod
    def to_vcsx(self) -> bytes:
        """Return LPEG VCSX data for this command."""

    @classmethod
    @abstractmethod
    def from_vcsx(cls: Type[_T], data: bytes) -> _T:
        """Construct command from LPEG VCSX data.

        Arguments:
            data: VCSX data.
        """

    @classmethod
    @abstractmethod
    def vcsx_size(cls) -> int:
        """Return size of VCSX command data in bytes."""


@dataclass
class VorzeVibrateCommand(BaseVorzeCommand, VibrateCommand):
    """Vorze vibration command.

    Attributes:
        speed: Vibration speed with a range of [0-100].
    """

    SPEED_DIVISOR: ClassVar[int] = 100

    speed: int

    @property
    def speeds(self) -> List[float]:
        """Return Buttplug VibrateCmd speeds for this command.

        Returns:
            List of speeds suitable for use with buttplug-py
            ``device.send_vibrate_cmd()``.
        """
        return [self.speed / self.SPEED_DIVISOR]

    @classmethod
    def from_speeds(
        cls, speeds: Union[List[SpeedSubcommand], List[float]]
    ) -> "VorzeVibrateCommand":
        """Return a command instance from Buttplug VibrateCmd speeds.

        Arguments:
            speeds: Buttplug VibrateCmd speeds list.

        Returns:
            New command instance.

        Raises:
            ValueError: Invalid speeds.
        """
        if not speeds:
            raise ValueError("Vibrate speeds cannot be empty.")
        speed = speeds[0]
        if isinstance(speed, SpeedSubcommand):
            speed = speed.speed
        return cls(round(speed * cls.SPEED_DIVISOR))

    def to_csv(self) -> Tuple[int]:
        """Return Vorze CSV row data for this command."""
        return (self.speed,)

    @classmethod
    def from_csv(
        cls: Type["VorzeVibrateCommand"], row: Iterable[Any]
    ) -> "VorzeVibrateCommand":
        """Construct command from a Vorze CSV row.

        Arguments:
            row: CSV row data.

        Returns:
            Vibration command.
        """
        (speed,) = row
        return cls(int(speed))

    @classmethod
    def vcsx_size(cls) -> int:
        """Return size of VCSX command data in bytes."""
        return 1

    def to_vcsx(self) -> bytes:
        """Return LPEG VCSX data for this command."""
        cmd = self.speed & 0x7F
        return bytes([cmd])

    @classmethod
    def from_vcsx(
        cls: Type["VorzeVibrateCommand"], data: bytes
    ) -> "VorzeVibrateCommand":
        """Construct command from LPEG VCSX data.

        Arguments:
            data: VCSX data.

        Returns:
            Rotation command.

        Raises:
            ValueError: Invalid VCSX data.
        """
        if not data:
            raise ValueError("Invalid VCSX data")
        cmd = data[0]
        speed = cmd & 0x7F
        return cls(speed)


@dataclass
class VorzeLinearCommand(BaseVorzeCommand, LinearPositionCommand):
    """Vorze linear movement (piston) command.

    Attributes:
        speed: Movement speed with a range of [0-100].
        position: Device position with a range of [0-200], where zero
            corresponds to the entrance end of the device.

    Note:
        Vorze linear movement commands direct the device to move to a
        designated position at a designated speed. Buttplug (and
        Funscript) linear movements direct the device to move to a
        designated position over some period of time. Due to this difference
        loss of resolution will occur when converting between Vorze and
        Buttplug commands.
    """

    POSITION_DIVISOR: ClassVar[int] = 200

    position: int
    speed: int

    def vectors(self, position: float) -> List[Tuple[int, float]]:
        """Return Buttplug LinearCmd vectors for this command.

        Arguments:
            position: Current device position in the range [0.0-1.0].

        Returns:
            List of vectors suitable for use with buttplug-py
            ``device.send_linear_cmd()``.
        """
        distance = abs(self.position - round(position * self.POSITION_DIVISOR))
        if distance == 0:
            duration = 0.0
        else:
            duration = (
                math.pow(1 / self.speed, 1 / 1.21)
                * 6658
                * distance
                / self.POSITION_DIVISOR
            )
        return [(round(duration), self.position / self.POSITION_DIVISOR)]

    @classmethod
    def from_vectors(
        cls,
        vectors: Union[List[LinearSubcommand], List[Tuple[int, float]]],
        position: float,
    ) -> "VorzeLinearCommand":
        """Return a command instance from Buttplug LinearCmd speeds.

        Arguments:
            vectors: Buttplug LinearCmd vectors list.
            position: Current device position in the range [0.0-1.0].

        Returns:
            New command instance.

        Raises:
            ValueError: Invalid vectors or position.
        """
        if not vectors:
            raise ValueError("Linear vectors cannot be empty.")
        if isinstance(vectors[0], LinearSubcommand):
            duration = vectors[0].duration
            new_position = vectors[0].position
        else:
            duration, new_position = vectors[0]
        curpos = round(position * cls.POSITION_DIVISOR)
        newpos = round(new_position * cls.POSITION_DIVISOR)
        distance = abs(newpos - curpos)
        if distance == 0:
            speed = 0
        else:
            # speed conversion from Buttplug device/protocol/vorze_sa.rs
            dur = duration * cls.POSITION_DIVISOR / distance
            speed = round(math.pow(dur / 6658, -1.21))
            if speed > 100:
                speed = 100
                logger.warning("Required movement exceeds max speed (using 100).")
            elif speed == 0:
                speed = 1
                logger.warning("Required movement below min speed (using 1).")
        return cls(newpos, speed)

    def to_csv(self) -> Tuple[int, int]:
        """Return Vorze CSV row data for this command."""
        return self.position, self.speed

    @classmethod
    def from_csv(
        cls: Type["VorzeLinearCommand"], row: Iterable[Any]
    ) -> "VorzeLinearCommand":
        """Construct command from a Vorze CSV row.

        Arguments:
            row: CSV row data.

        Returns:
            Vibration command.
        """
        (position, speed) = row
        return cls(int(position), int(speed))

    @classmethod
    def vcsx_size(cls) -> int:
        """Return size of VCSX command data in bytes."""
        return 3

    def to_vcsx(self) -> bytes:
        """Return LPEG VCSX data for this command."""
        cmd = [0, self.position & 0xFF, self.speed & 0xFF]
        return bytes(cmd)

    @classmethod
    def from_vcsx(cls: Type["VorzeLinearCommand"], data: bytes) -> "VorzeLinearCommand":
        """Construct command from LPEG VCSX data.

        Arguments:
            data: VCSX data.

        Returns:
            Rotation command.

        Raises:
            ValueError: Invalid VCSX data.
        """
        if len(data) < 3:
            raise ValueError("Invalid VCSX data")
        cmd = data[:3]
        position = cmd[1] & 0xFF
        speed = cmd[2] & 0xFF
        return cls(position, speed)


@dataclass
class VorzeRotateCommand(BaseVorzeCommand, RotateCommand):
    """Vorze rotation command.

    Attributes:
        speed: Rotation speed with a range of [0-100].
        clockwise: Rotation direction.
    """

    SPEED_DIVISOR: ClassVar[int] = 100

    speed: int
    clockwise: bool

    @property
    def rotations(self) -> List[Tuple[float, bool]]:
        """Return Buttplug RotateCmd rotations for this command.

        Returns:
            List of vectors suitable for use with buttplug-py
            ``device.send_rotate_cmd()``.
        """
        return [(self.speed / self.SPEED_DIVISOR, self.clockwise)]

    @classmethod
    def from_rotations(
        cls, rotations: Union[List[RotateSubcommand], List[Tuple[float, bool]]]
    ) -> "VorzeRotateCommand":
        """Return a command instance from Buttplug RotateCmd rotations.

        Arguments:
            rotations: Buttplug RotateCmd rotations list.

        Returns:
            New command instance.

        Raises:
            ValueError: Invalid rotations.
        """
        if not rotations:
            raise ValueError("Rotations cannot be empty.")
        rotation = rotations[0]
        if isinstance(rotation, RotateSubcommand):
            speed = rotation.speed
            clockwise = rotation.clockwise
        else:
            speed, clockwise = rotation
        return cls(round(speed * cls.SPEED_DIVISOR), clockwise)

    def to_csv(self) -> Tuple[int, int]:
        """Return Vorze CSV row data for this command."""
        return 0 if self.clockwise else 1, self.speed

    @classmethod
    def from_csv(
        cls: Type["VorzeRotateCommand"], row: Iterable[Any]
    ) -> "VorzeRotateCommand":
        """Construct command from a Vorze CSV row.

        Arguments:
            row: CSV row data.

        Returns:
            Rotation command.
        """
        direction, speed = row
        return cls(int(speed), int(direction) == 0)

    @classmethod
    def vcsx_size(cls) -> int:
        """Return size of VCSX command data in bytes."""
        return 1

    def to_vcsx(self) -> bytes:
        """Return LPEG VCSX data for this command."""
        cmd = (self.speed & 0x7F) | (0 if self.clockwise else 0x80)
        return bytes([cmd])

    @classmethod
    def from_vcsx(cls: Type["VorzeRotateCommand"], data: bytes) -> "VorzeRotateCommand":
        """Construct command from LPEG VCSX data.

        Arguments:
            data: VCSX data.

        Returns:
            Rotation command.

        Raises:
            ValueError: Invalid VCSX data.
        """
        if not data:
            raise ValueError("Invalid VCSX data")
        cmd = data[0]
        speed = cmd & 0x7F
        clockwise = cmd & 0x80 == 0
        return cls(speed, clockwise)



"""Generic utilities."""
from typing import Literal
from typing import Union

from typing_extensions import TypeAlias


_ByteOrder: TypeAlias = Union[Literal["little"], Literal["big"]]


def to_uint(data: bytes, byteorder: _ByteOrder = "big") -> int:
    """Convert bytes to unsigned integer."""
    return int.from_bytes(data, byteorder=byteorder, signed=False)


def to_u32(n: int, byteorder: _ByteOrder = "big") -> bytes:
    """Convert unsigned int to bytes."""
    return n.to_bytes(4, byteorder=byteorder, signed=False)



"""Script module."""
from .base import BaseScript
from .base import ScriptCommand
from .vcsx import VCSXCycloneScript
from .vcsx import VCSXOnaRhythmScript
from .vcsx import VCSXPistonScript
from .vcsx import VCSXScript
from .vorze import VorzeLinearScript
from .vorze import VorzeRotateScript
from .vorze import VorzeScript
from .vorze import VorzeVibrateScript


__all__ = [
    "BaseScript",
    "ScriptCommand",
    "VCSXCycloneScript",
    "VCSXOnaRhythmScript",
    "VCSXPistonScript",
    "VCSXScript",
    "VorzeLinearScript",
    "VorzeRotateScript",
    "VorzeScript",
    "VorzeVibrateScript",
]



"""Generic command module."""
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import List
from typing import Protocol
from typing import Tuple
from typing import Union

from buttplug.core import LinearSubcommand
from buttplug.core import RotateSubcommand
from buttplug.core import SpeedSubcommand


class BaseCommand(ABC):  # noqa: B024
    """Base a10sa-script command."""


class VibrateCommand(Protocol):
    """Vibration command protocol.

    Commands which implement VibrateCommand can be serialized to/from Buttplug
    VibrateCmd speeds.
    """

    @property
    @abstractmethod
    def speeds(self) -> List[float]:
        """Return Buttplug VibrateCmd speeds for this command.

        Returns:
            List of speeds suitable for use with buttplug-py
            ``device.send_vibrate_cmd()``.
        """

    @classmethod
    @abstractmethod
    def from_speeds(
        cls, speeds: Union[List[SpeedSubcommand], List[float]]
    ) -> "VibrateCommand":
        """Return a command instance from Buttplug VibrateCmd speeds.

        Arguments:
            speeds: Buttplug VibrateCmd speeds list.

        Returns:
            New command instance.
        """


@dataclass
class GenericVibrateCommand(BaseCommand, VibrateCommand):
    """Generic Buttplug-compliant vibration command.

    Attributes
        speed: Vibration speed with a range of [0.0-1.0].
    """

    speed: float

    @property
    def speeds(self) -> List[float]:
        """Return Buttplug VibrateCmd speeds for this command.

        Returns:
            List of speeds suitable for use with buttplug-py
            ``device.send_vibrate_cmd()``.
        """
        return [self.speed]

    @classmethod
    def from_speeds(
        cls, speeds: Union[List[SpeedSubcommand], List[float]]
    ) -> "GenericVibrateCommand":
        """Return a command instance from Buttplug VibrateCmd speeds.

        Arguments:
            speeds: Buttplug VibrateCmd speeds list.

        Returns:
            New command instance.

        Raises:
            ValueError: Invalid speeds.
        """
        if not speeds:
            raise ValueError("Vibrate speeds cannot be empty.")

        speed = speeds[0]
        if isinstance(speed, SpeedSubcommand):
            speed = speed.speed
        return cls(speed)


class LinearCommand(Protocol):
    """Linear command protocol.

    Commands which implement LinearCommand can be serialized to/from Buttplug
    LinearCmd vectors.
    """

    @property
    @abstractmethod
    def vectors(self) -> List[Tuple[int, float]]:
        """Return Buttplug LinearCmd vectors for this command.

        Returns:
            List of vectors suitable for use with buttplug-py
            ``device.send_linear_cmd()``.
        """

    @classmethod
    @abstractmethod
    def from_vectors(
        cls, vectors: Union[List[LinearSubcommand], List[Tuple[int, float]]]
    ) -> "LinearCommand":
        """Return a command instance from Buttplug LinearCmd speeds.

        Arguments:
            vectors: Buttplug LinearCmd vectors list.

        Returns:
            New command instance.
        """


class LinearPositionCommand(Protocol):
    """Alternate linear command protocol for devices which use absolute position.

    Commands which implement LinearPosition Command can be serialized to/from
    Buttplug LinearCmd vectors as long as the current device position is known.
    """

    @abstractmethod
    def vectors(self, position: float) -> List[Tuple[int, float]]:
        """Return Buttplug LinearCmd vectors for this command.

        Arguments:
            position: Current device position in the range [0.0-1.0].

        Returns:
            List of vectors suitable for use with buttplug-py
            ``device.send_linear_cmd()``.
        """

    @classmethod
    @abstractmethod
    def from_vectors(
        cls,
        vectors: Union[List[LinearSubcommand], List[Tuple[int, float]]],
        position: float,
    ) -> "LinearPositionCommand":
        """Return a command instance from Buttplug LinearCmd speeds.

        Arguments:
            vectors: Buttplug LinearCmd vectors list.
            position: Current device position in the range [0.0-1.0].

        Returns:
            New command instance.
        """


@dataclass
class GenericLinearCommand(BaseCommand, LinearCommand):
    """Generic linear movement command.

    Attributes:
        duration: Movement time in milliseconds.
        position: Target position with a range of [0.0-1.0].
    """

    duration: int
    position: float

    @property
    def vectors(self) -> List[Tuple[int, float]]:
        """Return Buttplug LinearCmd vectors for this command.

        Returns:
            List of vectors suitable for use with buttplug-py
            ``device.send_linear_cmd()``.
        """
        return [(self.duration, self.position)]

    @classmethod
    def from_vectors(
        cls, vectors: Union[List[LinearSubcommand], List[Tuple[int, float]]]
    ) -> "GenericLinearCommand":
        """Return a command instance from Buttplug LinearCmd speeds.

        Arguments:
            vectors: Buttplug LinearCmd vectors list.

        Returns:
            New command instance.

        Raises:
            ValueError: Invalid vectors.
        """
        if not vectors:
            raise ValueError("Linear vectors cannot be empty.")
        vector = vectors[0]
        if isinstance(vector, LinearSubcommand):
            vector = (vector.duration, vector.position)
        return cls(*vector)


class RotateCommand(Protocol):
    """Rotate command protocol.

    Commands which implement RotateCommand can be serialized to/from Buttplug
    RotateCmd rotations.
    """

    @property
    @abstractmethod
    def rotations(self) -> List[Tuple[float, bool]]:
        """Return Buttplug RotateCmd rotations for this command.

        Returns:
            List of vectors suitable for use with buttplug-py
            ``device.send_rotate_cmd()``.
        """

    @classmethod
    @abstractmethod
    def from_rotations(
        cls, rotations: Union[List[RotateSubcommand], List[Tuple[float, bool]]]
    ) -> "RotateCommand":
        """Return a command instance from Buttplug RotateCmd rotations.

        Arguments:
            rotations: Buttplug RotateCmd rotations list.

        Returns:
            New command instance.
        """


@dataclass
class GenericRotateCommand(BaseCommand, RotateCommand):
    """Generic rotation command.

    Attributes:
        speed: Rotation speed with a range of [0.0-1.0].
        clockwise: Direction of rotation.
    """

    speed: float
    clockwise: bool

    @property
    def rotations(self) -> List[Tuple[float, bool]]:
        """Return Buttplug RotateCmd rotations for this command.

        Returns:
            List of vectors suitable for use with buttplug-py
            ``device.send_rotate_cmd()``.
        """
        return [(self.speed, self.clockwise)]

    @classmethod
    def from_rotations(
        cls, rotations: Union[List[RotateSubcommand], List[Tuple[float, bool]]]
    ) -> "GenericRotateCommand":
        """Return a command instance from Buttplug RotateCmd rotations.

        Arguments:
            rotations: Buttplug RotateCmd rotations list.

        Returns:
            New command instance.

        Raises:
            ValueError: Invalid rotations.
        """
        if not rotations:
            raise ValueError("Rotations cannot be empty.")
        rotation = rotations[0]
        if isinstance(rotation, RotateSubcommand):
            rotation = (rotation.speed, rotation.clockwise)
        return cls(*rotation)



"""LPEG/Afesta VCSX script module."""
from abc import abstractmethod
from typing import Any
from typing import BinaryIO
from typing import ClassVar
from typing import Iterable
from typing import Optional
from typing import Type

from ..command.vorze import VorzeLinearCommand
from ..command.vorze import VorzeRotateCommand
from ..command.vorze import VorzeVibrateCommand
from ..exceptions import ParseError
from ..utils import to_u32
from ..utils import to_uint
from .base import SerializableScript
from .vorze import _SC
from .vorze import _T
from .vorze import VorzeScriptCommand


class VCSXScript(SerializableScript[_T]):
    """Generic LPEG VCSX script.

    Note:
        LPEG VCSX is a container for Vorze's binary BLE command format.
    """

    VCSX_MAGIC: ClassVar[bytes]
    VCSX_DEFAULT_VERSION: ClassVar[bytes]

    def __init__(
        self,
        commands: Optional[Iterable[_SC[_T]]] = None,
        version: Optional[bytes] = None,
        **kwargs: Any,
    ):
        """Construct a script.

        Arguments:
            commands: Initial set of commands for this script.
            version: VCSX version.
            kwargs: Additional kwargs.
        """
        super().__init__(commands, **kwargs)
        self.version = version or self.VCSX_DEFAULT_VERSION

    @classmethod
    @abstractmethod
    def _command_cls(cls) -> Type[_T]:
        """Return command class for this script."""

    def dump(self, fp: BinaryIO) -> None:
        """Serialize script to file.

        Arguments:
            fp: A file-like object opened for writing.
        """
        fp.write(self.VCSX_MAGIC + self.version + to_u32(len(self)))
        for cmd in self.commands:
            fp.write(to_u32(cmd.offset) + cmd.cmd.to_vcsx())

    @classmethod
    def load(cls, fp: BinaryIO) -> "VCSXScript[_T]":
        """Deserialize script from file.

        Arguments:
            fp: A file-like object opened for reading.

        Returns:
            Loaded command script.

        Raises:
            ParseError: A CSV parsing error occured.
        """
        try:
            magic = fp.read(len(cls.VCSX_MAGIC))
            if magic != cls.VCSX_MAGIC:
                raise ParseError("Invalid VCSX header")
            version = fp.read(len(cls.VCSX_DEFAULT_VERSION))
            size = to_uint(fp.read(4))
            cmd_cls = cls._command_cls()
            return cls(
                (
                    VorzeScriptCommand(
                        to_uint(fp.read(4)),
                        cmd_cls.from_vcsx(fp.read(cmd_cls.vcsx_size())),
                    )
                    for i in range(size)
                ),
                version=version,
            )
        except (OSError, ValueError) as e:
            raise ParseError("Failed to parse file as VCSX data.") from e


class VCSXOnaRhythmScript(VCSXScript[VorzeVibrateCommand]):
    """VCSX OnaRhythm script."""

    VCSX_MAGIC = b"VCSX\x01Vorze_OnaRhythm\x00"
    VCSX_DEFAULT_VERSION = b"\x01\x42"

    @classmethod
    def _command_cls(cls) -> Type[VorzeVibrateCommand]:
        """Return command class for this script."""
        return VorzeVibrateCommand


class VCSXPistonScript(VCSXScript[VorzeLinearCommand]):
    """VCSX Piston script."""

    VCSX_MAGIC = b"VCSX\x01Vorze_Piston\x00"
    VCSX_DEFAULT_VERSION = b"\x02\x57\x42"

    @classmethod
    def _command_cls(cls) -> Type[VorzeLinearCommand]:
        """Return command class for this script."""
        return VorzeLinearCommand


class VCSXCycloneScript(VCSXScript[VorzeRotateCommand]):
    """VCSX Cyclone SA script."""

    VCSX_MAGIC = b"VCSX\x01Vorze_CycloneSA\x00"
    VCSX_DEFAULT_VERSION = b"\x01\x42"

    @classmethod
    def _command_cls(cls) -> Type[VorzeRotateCommand]:
        """Return command class for this script."""
        return VorzeRotateCommand



"""Command-line interface."""
import click


@click.command()
@click.version_option()
def main() -> None:
    """A10SA Script."""


if __name__ == "__main__":
    main(prog_name="a10sa-script")  # pragma: no cover



"""Base script module."""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import BinaryIO
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import MutableSequence
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import overload

from sortedcontainers import SortedList
from typing_extensions import TypeAlias

from ..command import BaseCommand


_T = TypeVar("_T", bound=BaseCommand)


@dataclass
class ScriptCommand(Generic[_T]):
    """Timed script command.

    Attributes:
        offset: Offset from start of script in milliseconds.
        cmd: Command.
    """

    offset: int
    cmd: _T


_SC: TypeAlias = ScriptCommand


class BaseScript(MutableSequence[_SC[_T]]):
    """Base command script."""

    def __init__(
        self,
        commands: Optional[Iterable[_SC[_T]]] = None,
        **kwargs: Any,
    ):
        """Construct a script.

        Arguments:
            commands: Initial set of commands for this script.
            kwargs: Additional kwargs.
        """
        self.commands = SortedList(commands, key=lambda cmd: cmd.offset)

    @overload
    def __getitem__(self, key: int) -> _SC[_T]:
        ...

    @overload
    def __getitem__(self, key: slice) -> MutableSequence[_SC[_T]]:
        ...

    def __getitem__(
        self, key: Union[int, slice]
    ) -> Union[_SC[_T], MutableSequence[_SC[_T]]]:
        return self.commands[key]

    def __setitem__(
        self,
        key: Union[int, slice],
        value: Union[_SC[_T], Iterable[_SC[_T]]],
    ) -> None:
        raise NotImplementedError

    def __delitem__(self, key: Union[int, slice]) -> None:
        del self.commands[key]

    def __len__(self) -> int:
        return len(self.commands)

    def __iter__(self) -> Iterator[_SC[_T]]:
        return iter(self.commands)

    def __reversed__(self) -> Iterator[_SC[_T]]:
        return reversed(self.commands)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, BaseScript) and self.commands == other.commands

    def insert(self, index: int, command: _SC[_T]) -> None:
        """Raise not-implemented error (use add())."""
        raise NotImplementedError

    def add(self, command: _SC[_T]) -> None:
        """Add a command to this script.

        Commands will be inserted in the proper location based on time offset.

        Arguments:
            command: Command to insert.
        """
        self.commands.add(command)

    def reverse(self) -> None:
        """Raise not-implemented error.

        In place reversal is not supported (script will always be sorted).
        """
        raise NotImplementedError

    def seek_iter(self, offset: int) -> Iterator[_SC[_T]]:
        """Return an iterator for commands starting at the specified offset.

        Arguments:
            offset: Time offset in milliseconds.

        Returns:
            Command iterator.
        """
        return self.commands.irange_key(offset)


class SerializableScript(BaseScript[_T]):
    """A command script which can be serialized."""

    @abstractmethod
    def dump(self, fp: BinaryIO) -> None:
        """Serialize script to file.

        Arguments:
            fp: A file-like object opened for writing.
        """

    @classmethod
    @abstractmethod
    def load(cls, fp: BinaryIO) -> "SerializableScript[_T]":
        """Deserialize script from file.

        Arguments:
            fp: A file-like object opened for reading.
        """



"""Vorze CSV script module."""
import csv
import io
from abc import abstractmethod
from dataclasses import dataclass
from typing import BinaryIO
from typing import Type
from typing import TypeVar

from typing_extensions import TypeAlias

from ..command.vorze import BaseVorzeCommand
from ..command.vorze import VorzeLinearCommand
from ..command.vorze import VorzeRotateCommand
from ..command.vorze import VorzeVibrateCommand
from ..exceptions import ParseError
from .base import ScriptCommand
from .base import SerializableScript


_T = TypeVar("_T", bound=BaseVorzeCommand)


@dataclass
class VorzeScriptCommand(ScriptCommand[_T]):
    """Timed Vorze script command."""


_SC: TypeAlias = VorzeScriptCommand


class VorzeScript(SerializableScript[_T]):
    """Generic Vorze CSV script.

    Note:
        Vorze CSVs time offsets are integers in tenths of a second. Internally
        we store offsets in milliseconds so loss of resolution will occur when
        converting to CSV.
    """

    OFFSET_DENOM = 100

    @classmethod
    @abstractmethod
    def _command_cls(cls) -> Type[_T]:
        """Return command class for this script."""

    def dump(self, fp: BinaryIO) -> None:
        """Serialize script to file.

        Arguments:
            fp: A file-like object opened for writing.
        """
        text_fp = io.TextIOWrapper(fp, newline="")
        try:
            writer = csv.writer(text_fp)
            writer.writerows(
                (self.offset_from_ms(cmd.offset),) + tuple(cmd.cmd.to_csv())
                for cmd in self.commands
            )
        finally:
            text_fp.detach()

    @classmethod
    def load(cls, fp: BinaryIO) -> "VorzeScript[_T]":
        """Deserialize script from file.

        Arguments:
            fp: A file-like object opened for reading.

        Returns:
            Loaded command script.

        Raises:
            ParseError: A CSV parsing error occured.
        """
        try:
            reader = csv.reader(io.TextIOWrapper(fp, newline=""))
            return cls(
                VorzeScriptCommand(
                    cls.offset_to_ms(int(row[0])),
                    cls._command_cls().from_csv(row[1:]),
                )
                for row in reader
            )
        except (csv.Error, ValueError) as e:
            raise ParseError("Failed to parse file as Vorze CSV.") from e

    @classmethod
    def offset_from_ms(cls, offset_ms: int) -> int:
        """Convert millisecond time offset to Vorze CSV offsets."""
        return round(offset_ms / cls.OFFSET_DENOM)

    @classmethod
    def offset_to_ms(cls, offset_csv: int) -> int:
        """Convert Vorze CSV time offsets to millisecond offsets."""
        return offset_csv * cls.OFFSET_DENOM


class VorzeVibrateScript(VorzeScript[VorzeVibrateCommand]):
    """Vorze vibration commands script.

    Note:
        Script format for Vorze OnaRhythm devices (i.e Rocket+1D).
    """

    @classmethod
    def _command_cls(cls) -> Type[VorzeVibrateCommand]:
        """Return command class for this script."""
        return VorzeVibrateCommand


class VorzeLinearScript(VorzeScript[VorzeLinearCommand]):
    """Vorze linear commands script.

    Note:
        Script format for Piston devices.
    """

    @classmethod
    def _command_cls(cls) -> Type[VorzeLinearCommand]:
        """Return command class for this script."""
        return VorzeLinearCommand


class VorzeRotateScript(VorzeScript[VorzeRotateCommand]):
    """Vorze rotation commands script.

    Note:
        Shared script format for both Cyclone and UFO devices.
    """

    @classmethod
    def _command_cls(cls) -> Type[VorzeRotateCommand]:
        """Return command class for this script."""
        return VorzeRotateCommand



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['a10sa_script', 'a10sa_script.command', 'a10sa_script.script']

package_data = \
{'': ['*']}

install_requires = \
['buttplug>=0.3.0,<0.4.0',
 'click>=8.0.1',
 'loguru>=0.6.0,<0.7.0',
 'sortedcontainers>=2.4.0,<3.0.0',
 'typing-extensions>=4.0.1,<5.0.0']

entry_points = \
{'console_scripts': ['a10sa-script = a10sa_script.__main__:main']}

setup_kwargs = {
    'name': 'a10sa-script',
    'version': '0.0.4',
    'description': 'A10SA Script',
    'long_description': 'A10SA Script\n============\n\n|PyPI| |Status| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/a10sa-script.svg\n   :target: https://pypi.org/project/a10sa-script/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/a10sa-script.svg\n   :target: https://pypi.org/project/a10sa-script/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/a10sa-script\n   :target: https://pypi.org/project/a10sa-script\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/a10sa-script\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/a10sa-script/latest.svg?label=Read%20the%20Docs\n   :target: https://a10sa-script.readthedocs.io/\n   :alt: Read the documentation at https://a10sa-script.readthedocs.io/\n.. |Tests| image:: https://github.com/bhrevol/a10sa-script/workflows/Tests/badge.svg\n   :target: https://github.com/bhrevol/a10sa-script/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/bhrevol/a10sa-script/branch/main/graph/badge.svg\n   :target: https://app.codecov.io/gh/bhrevol/a10sa-script\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\nLibrary for manipulating Vorze A10 sex toy scripts.\n\n\nFeatures\n--------\n\n* Read/Write/Convert supported script formats.\n* Export scripts as `Buttplug Protocol`_ command sequences.\n\n.. _Buttplug Protocol: https://buttplug.io/\n\nSupported Formats\n-----------------\n\n* Vorze CSV\n* Afesta/LPEG VCSX (``.bin``)\n\n\nRequirements\n------------\n\n* Python 3.8+\n\n\nInstallation\n------------\n\nYou can install *A10SA Script* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install a10sa-script\n\n\nUsage\n-----\n\nConvert VCSX ``Vorze_CycloneSA.bin`` to ``script_cyclone.csv``:\n\n.. code:: py\n\n    >>> from a10sa_script.script import VCSXCycloneScript, VorzeRotateScript\n    >>> with open("Vorze_CycloneSA.bin", "rb") as f:\n    ...     vcsx = VCSXCycloneScript.load(f)\n    >>> with open("script_cyclone.csv", "wb") as f:\n    ...     VorzeRotateScript(vcsx.commands).dump(f)\n\nPlease see the `Command-line Reference <Usage_>`_ for details.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*A10SA Script* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_\'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/bhrevol/a10sa-script/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: https://a10sa-script.readthedocs.io/en/latest/contributing.html\n.. _Usage: https://a10sa-script.readthedocs.io/en/latest/usage.html\n',
    'author': 'byeonhyeok',
    'author_email': 'bhrevol@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bhrevol/a10sa-script',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Device command module."""
from .base import BaseCommand
from .base import GenericLinearCommand
from .base import GenericRotateCommand
from .base import GenericVibrateCommand
from .base import LinearCommand
from .base import LinearPositionCommand
from .base import RotateCommand
from .base import VibrateCommand
from .vorze import VorzeLinearCommand
from .vorze import VorzeRotateCommand
from .vorze import VorzeVibrateCommand


__all__ = [
    "BaseCommand",
    "GenericLinearCommand",
    "GenericRotateCommand",
    "GenericVibrateCommand",
    "LinearCommand",
    "LinearPositionCommand",
    "RotateCommand",
    "VibrateCommand",
    "VorzeLinearCommand",
    "VorzeRotateCommand",
    "VorzeVibrateCommand",
]
